/* (c) 1996-2001, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* Tagfile Identification Database Interface
 *
 * This module supports using tagfiles as an identification database.  This
 * is a file named 'id' that sits in the user's home directory and has lines
 * like 'fullname=(Jan Wolter)'.  The directory path isn't stored in the
 * tagfile, because you need it before you can access the tagfile.  Instead
 * we generate the directory name from the login name.
 */


#include "backtalk.h"
#include "authident.h"

#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif

#include "set_uid.h"
#include "str.h"
#include "udb.h"
#include "lock.h"
#include "entropy.h"

#ifdef DMALLOC
#include "dmalloc.h"
#endif

char *showopt_ident_module= "ident_tagfile";

/* SAVE_TAG - Write the given variable/value pair into a tag file, replacing
 * the old value, if there was one, or appending the value.  The value is
 * not changed if it was protected, in which case a non-zero value is returned.
 * This always takes four arguments, but the last may be either an integer or
 * a string depending on the type.
 */

#if __STDC__
int save_tag(char *fname, int amadm, char *varname, int type, ...)
#else
int save_tag(fname, amadm, varname, type, va_alist)
    char *fname;
    int amadm;
    char *varname;
    int type;
    va_dcl
#endif
{
    va_list ap;
    char *bf, *val, *end, *var;
    FILE *fp;
    int i, protect, done= 0, rc= 0;

    bf= read_file(fname);
    VA_START(ap, type);

    /* Truncate and prepare to rewrite */
    if ((fp= fopen(fname,"w+")) == NULL)
	die("Cannot open file %s to write",fname);
    lock_exclusive(fileno(fp),fname);

    /* Process each line of saved file contents, writing back or replacing */
    for (var= bf; var != NULL && *var != '\0'; var= end+1)
    {
	/* Parse file line */
	if (protect= (*var == '*')) var++;
	val= strchr(var,'=');
	end= strchr(var,'\n');
	if (val != NULL && end != NULL && val > var && end > val)
	{
	    *(val++)= '\0';
	    *end= '\0';
	    if (!done && !strcmp(var,varname))
	    {
		done= 1;
		if (protect && !amadm)
		    rc= 1;
		else
		{
		    if (type == TK_INTEGER)
			fprintf(fp,"%s%s=%d\n", protect?"*":"",
			    var, va_arg(ap, int));
		    else
		    {
			if (protect) fputc('*',fp);
			wrtstrtag(fp,var,va_arg(ap, char *));
		    }
		    continue;
		}
	     }
	     fprintf(fp,"%s=%s\n", var-protect, val);
	}
    }

    /* Process the variable if not previously in file */
    if (!done)
    {
	if (type == TK_INTEGER)
	    fprintf(fp,"%s=%d\n", varname, va_arg(ap, int));
	else
	    wrtstrtag(fp, varname, va_arg(ap, char *));
    }

    fclose(fp);
    unlock(fname);
    if (bf != NULL) free(bf);
    va_end(ap);
    return rc;
}


/* IDENT_GETPWNAM - Get the database entry for the named user.
 */

#ifdef HAVEIDENT_GETPWNAM
struct btpasswd *ident_getpwnam(char *nam)
{
    static struct btpasswd pw;
    static char stbf[BFSZ];
    char fname[BFSZ+1];
    char bf[BFSZ+1], *dir, *var, *val, *end, *stp;
    FILE *fp;
    int ch, gotuid, gotgid;

    pw.name= nam;

    pw.email= pw.gecos= pw.valcode= NULL;
    pw.priv= pw.epriv= -1;
    gotuid= gotgid= FALSE;
    stp= stbf;

    /* Construct the directory name */
    if (nam == NULL || (dir= homedirpath(nam)) == NULL)
	return NULL;
    pw.dir= dir;

    /* Open tag file */
    sprintf(fname,"%s/"IDENT_FILE,dir);
    if ((fp= fopen(fname,"r")) == NULL)
	    return NULL;
    lock_shared(fileno(fp),fname);

    
    /* Read through tag file */
    while (fgets(bf,BFSZ,fp) != NULL)
    {
	/* Parse apart the line */
	var= bf+ (bf[0] == '*');
	if ((val= strchr(var, '=')) == NULL)
	{
	    fclose(fp);
	    unlock(fname);
	    return NULL;
	}
	*(val++)= '\0';
	end= val + strlen(val) - 1;
	if (*end == '\n')
	    *(end--)= '\0';
	else
	    /* skip over rest of over-long line */
	    while ((ch= fgetc(fp)) != '\n' && ch != EOF)
		;

	if (!strcmp(var, "uid"))
	{
	    if (!isascii(*val) || !isdigit(*val)) goto error;
	    pw.uid= atoi(val);
	    gotuid= TRUE;
	}
	else if (!strcmp(var, "gid"))
	{
	    if (!isascii(*val) || !isdigit(*val)) goto error;
	    pw.gid= atoi(val);
	    gotgid= TRUE;
	}
	else if (!strcmp(var, "fullname"))
	{
	    /* We recognize both curly and round braces so both
	     * old and new file format works
	     */
	    if (*val != '(' && *val != '{') goto error;
	    if (*end == ')' || *end == '}') *end= '\0';
	    pw.gecos= stp;
	    stp= unesctag(val+1, pw.gecos, BFSZ - (stp - stbf), 0) + 1;
	    if (stp == NULL) goto error;
	}
	else if (!strcmp(var, "email"))
	{
	    if (*val != '(') goto error;
	    if (*end == ')') *end= '\0';
	    pw.email= stp;
	    stp= unesctag(val+1, pw.email, BFSZ - (stp - stbf), 0) + 1;
	    if (stp == NULL) goto error;
	}
	else if (!strcmp(var, "valcode"))
	{
	    if (*val != '(') goto error;
	    if (*end == ')') *end= '\0';
	    pw.valcode= stp;
	    stp= unesctag(val+1, pw.valcode, BFSZ - (stp - stbf), 0) + 1;
	    if (stp == NULL) goto error;
	}
	else if (!strcmp(var, "privacy"))
	{
	    if (!isascii(*val) || !isdigit(*val)) goto error;
	    pw.priv= atoi(val);
	}
	else if (!strcmp(var, "eprivacy"))
	{
	    if (!isascii(*val) || !isdigit(*val)) goto error;
	    pw.epriv= atoi(val);
	}
    }

    if (gotuid && gotgid)
    {
	fclose(fp);
	unlock(fname);
	return &pw;
    }

error:
    fclose(fp);
    unlock(fname);
    return NULL;
}
#endif


/* IDENT_NEWFNAME - change the user's fullname to the given value.
 */

#ifdef HAVEIDENT_NEWFNAME
void ident_newfname(struct btpasswd *pw, char *newfname)
{
    char fname[BFSZ];

    /* Construct name of ident file */
    sprintf(fname,"%s/"IDENT_FILE,pw->dir);

    flushusercache();

    if (save_tag(fname, 0, "fullname", TK_STRING, newfname))
	die("Protected fullname field in %s",fname);
}
#endif


/* IDENT_NEWGID - change the user's groupid to the given value.
 */

#ifdef HAVEIDENT_NEWGID
void ident_newgid(struct btpasswd *pw, gid_t newgid)
{
    char fname[BFSZ];

    /* Construct name of ident file */
    sprintf(fname,"%s/"IDENT_FILE,pw->dir);

    flushusercache();

    save_tag(fname, 1, "gid", TK_INTEGER, newgid);
}
#endif

/* IDENT_NEWEMAIL - change the user's email address to the given value.
*  */

#ifdef HAVEIDENT_NEWEMAIL
void ident_newemail(struct btpasswd *pw, char *newemail, int newvalid)
{
    char fname[BFSZ];
    char *valcode= NULL;

    /* Construct name of ident file */
    sprintf(fname,"%s/"IDENT_FILE,pw->dir);

    flushusercache();

    if (newemail[0] != '\0' && !newvalid)
	valcode= get_noise();

    /* A bit inefficient here - be nicer to do both in one pass */
    save_tag(fname, 1, "email", TK_STRING, newemail);
    save_tag(fname, 1, "valcode", TK_STRING, valcode == NULL ? "" : valcode);
}
#endif


#ifdef HAVEIDENT_NEWPRIV
void ident_newpriv(struct btpasswd *pw, int priv, int epriv)
{
    char fname[BFSZ];

    /* Construct name of ident file */
    sprintf(fname,"%s/"IDENT_FILE,pw->dir);

    flushusercache();

    /* A bit inefficient here - be nicer to do both in one pass */
    save_tag(fname, 1, "privacy", TK_INTEGER, priv);
    save_tag(fname, 1, "eprivacy", TK_INTEGER, epriv);
}
#endif


/* IDENT_ADD - Add an entry to the user database describing a user with
 * the given login, uid, gid, password, fullname, and directory.
 */

#ifdef HAVEIDENT_ADD
void ident_add(char *lid, uid_t uid, gid_t gid, char *fn, char *dir,
	char *email, int valid, int priv, int epriv)
{
    char fname[BFSZ];
    char *valcode= NULL;
    FILE *fp;

    if (email != NULL && email[0] != '\0' && !valid)
	valcode= get_noise();

    sprintf(fname,"%s/"IDENT_FILE,dir);
    if ((fp= fopen(fname,"w")) == NULL)
	die("could not create ident file %s",fname);
    lock_exclusive(fileno(fp),fname);
    fprintf(fp,"*uid=%d\n*gid=%d\n", uid, gid);
    if (fn != NULL) wrtstrtag(fp,"fullname",fn);
    if (email != NULL) wrtstrtag(fp,"email",email);
    if (valcode != NULL) wrtstrtag(fp,"valcode",valcode);
    if (priv != -1) fprintf(fp,"privacy=%d\n", priv);
    if (epriv != -1) fprintf(fp,"eprivacy=%d\n", epriv);
    fclose(fp);
    unlock(fname);
}
#endif


/* IDENT_DEL - Delete the named user from the ident database.  Actually, there
 * is nothing to do here, since we will always be deleting the directory
 * next, and that will take care of everything.
 */

#ifdef HAVEIDENT_DEL
void ident_del(char *name) { }
#endif
